#include "command.h"

#include "config.h"
#include "led.h"
#include "usart.h"

/*
 * The mode this module is running in.
 */
static mode_t my_mode = NORMAL_MODE;

/*
 * The commands that may arrive on the USART.
 */
typedef enum {
	// normal mode
	CMD_BROADCAST = 0xfd,
	CMD_STROBE = 0xfe,
	// config mode
	CMD_SET_CORRECTION = 0x01,
	CMD_SET_PERMUTATION = 0x02,
	CMD_SET_COLOR = 0x03,
	CMD_SET_HEAT_LIMIT = 0x04,
	CMD_RESET_CORRECTION = 0x05,
	CMD_SET_ADDRESS = 0x06,
	CMD_SET_BRIGHTNESS = 0x07,
	CMD_DUMP_CONFIG = 0x08,
	CMD_SAVE_CONFIG = 0xfe
} command_t;

/*
 * The expected length of each of the config mode commands.
 * Note that SAVE_CONFIG is not included.
 */
static int command_len[] = {
	0, // command 0 is invalid
	4, // SET_CORRECTION
	3, // SET_PERMUTATION
	3, // SET_COLOR
	3, // SET_HEAT_LIMIT
	1, // RESET_CORRECTION
	2, // SET_ADDRESS
	MODULE_LENGTH + 1, // SET_BRIGHTNESS
	1, // DUMP_CONFIG
};

/*
 * Returns the length of data (including the command) to be
 * expected for a given command code. If the command code is invalid,
 * 0 is returned.
 *
 * set_mode must be called before this.
 */
int command_length(char command_code) {
	if (my_mode == NORMAL_MODE) {
		// All commands have module length + 1,
		// Strobe has length 1.
		if (command_code == CMD_STROBE) {
			return 1;
		} else {
			return MODULE_LENGTH + 1;
		}
	} else {
		if (command_code == CMD_SAVE_CONFIG) {
			return 1;
		} else if (1 <= command_code &&
			   command_code <= 8) {
			return command_len[(int) command_code];
		} else {
			return 0;
		}
	}
}

/*
 * The USART filter function for normal mode.
 *
 * This function accepts the two addresses of the module, the broadcast
 * address and all special commands.
 */
static uint8_t normal_mode_filter(uint8_t command) {
#ifdef TRACE_USART
	debug_string("t");
	debug_putchar(config.my_address);
#endif
	
	return command == config.my_address ||
		command == CMD_BROADCAST ||
		command == CMD_STROBE;
}

/*
 * The USART filter function for config mode.
 *
 * This function accepts all comamnds.
 */
static uint8_t config_mode_filter(uint8_t command) {
	return 1;
}

/*
 * Sets the operating mode to normal or config mode, therefore deciding how
 * the commands are interpreted.
 *
 * The 'mode' argument should be NORMAL_MODE or CONFIG_MODE.
 */
void set_mode(mode_t mode) {
	if (mode == NORMAL_MODE) {
		usart_set_filter(normal_mode_filter);
	} else {
		usart_set_filter(config_mode_filter);
	}
	my_mode = mode;
}

/*
 * Returns the operating mode.
 */
mode_t get_mode() {
	return my_mode;
}

/*
 * Runs a "set LEDs" command. The expected format is:
 * <Module address> <LED0 brightness> <LED1 brightness> ...
 */
static error_t run_set_leds(uint8_t *led_values, uint8_t length) {
#ifdef TRACE_COMMANDS
	debug_string("set");
	debug_putchar(length);
	debug_write((char*) led_values, length);
#endif

	error_t error;

	if (MODULE_LENGTH < length) {
		length = MODULE_LENGTH;
	}

	for (uint8_t i = 0; i < length; i++) {
		error = led_set_brightness(config.physical_led[i], led_values[i]);

		if (error) return error;
	}

	if (get_mode() == CONFIG_MODE) {
		led_send_frame();
	}

	return E_SUCCESS;
}

/*
 * Runs a "set whitepoint correction" command. The expected format is:
 * <LED index> <correction factor>
 */
static error_t run_set_correction(uint8_t led_index, uint16_t factor) {
#ifdef TRACE_COMMANDS
	debug_string("corr");
	debug_putchar(led_index);
	debug_write(&factor, 2);
#endif

	if (led_index >= MODULE_LENGTH) {
		return E_INDEXRANGE;
	}

	config.white_correction[led_index] = factor;

	// Redisplay all LEDs so that the user can see the effect
	// of his calibration.
	led_send_frame();

	return E_SUCCESS;
}

/*
 * Runs a "reset whitepoint correction" command.
 */
static error_t run_reset_correction() {
#ifdef TRACE_COMMANDS
	debug_string("reset");
#endif

	for (uint8_t i = 0; i < MODULE_LENGTH; i++) {
		config.white_correction[i] = 0xff;
	}

	// Redisplay all LEDs so that the user can see the effect
	// of his calibration.
	led_send_frame();

	return E_SUCCESS;
}

/*
 * Runs a "set LED permutation" command.
 */
static error_t run_set_permutation(uint8_t physical, uint8_t logical) {
#ifdef TRACE_COMMANDS
	debug_string("perm");
	debug_putchar(physical);
	debug_putchar(logical);
#endif
	config.physical_led[logical] = physical;
	return E_SUCCESS;
}

/*
 * Runs a "set LED color" command.
 */
static error_t run_set_color(uint8_t led_index, color_t color) {
#ifdef TRACE_COMMANDS
	debug_string("scol");
	debug_putchar(led_index);
	debug_putchar(color);
#endif

	if (color == RED ||
	    color == GREEN ||
	    color == BLUE ||
	    color == WHITE) {
		config.led_color[led_index] = color;
		return E_SUCCESS;
	} else {
		return E_WRONGCOMMAND;
	}
}

/*
 * Runs a "set heat limit" command.
 */
static error_t run_set_heat_limit(uint8_t sensor, uint8_t limit) {
#ifdef TRACE_COMMANDS
	debug_string("sheat");
	debug_putchar(sensor);
	debug_putchar(limit);
#endif

	if (sensor < HEAT_SENSOR_LEN) {
		config.heat_limits[sensor] = limit;
		return E_SUCCESS;
	} else {
		return E_INDEXRANGE;
	}
}

/*
 * Runs a "set address" command.
 */
static error_t run_set_address(uint8_t address) {
#ifdef TRACE_COMMANDS
	debug_string("addr");
	debug_putchar(address);
#endif

	if (address != CMD_BROADCAST &&
	    address != CMD_STROBE) {
		config.my_address = address;
		return E_SUCCESS;
	} else {
		return E_INDEXRANGE;
	}
}

/*
 * Runs a "dump config" command.
 */
static error_t run_dump_config() {
	debug_write((char*) &config, sizeof(config));
	return E_SUCCESS;
}

/*
 * Runs the command pointed to by 'command', interpreting it in normal mode
 *
 * Returns an error/success code.
 */
static error_t run_command_normal(uint8_t *command) {

	if (command[0] == config.my_address ||
	    command[0] == CMD_BROADCAST) {
		// The first byte is the addresss. Drop it.
		return run_set_leds(command + 1, MODULE_LENGTH);
	} else if (command[0] == CMD_STROBE) {
		return led_send_frame();
	}

	// else
	return E_WRONGCOMMAND;
}

/*
 * Runs the command pointed to by 'command', interpreting it in config mode
 *
 * Returns an error/success code.
 */
static error_t run_command_config(uint8_t *command) {
	uint16_t wp_factor;

	switch(command[0]) {

	case CMD_SET_CORRECTION:
		wp_factor = command[2];
		wp_factor = wp_factor << 8;
		wp_factor += command[3];

		debug_putchar(command[2]);
		debug_putchar(command[3]);
		
		return run_set_correction(command[1], wp_factor);
	case CMD_SET_PERMUTATION:
		return run_set_permutation(command[1], command[2]);
	case CMD_SET_COLOR:
		return run_set_color(command[1], command[2]);
	case CMD_SET_HEAT_LIMIT:
		return run_set_heat_limit(command[1], command[2]);
	case CMD_RESET_CORRECTION:
		run_reset_correction();
		return E_SUCCESS;
	case CMD_SET_ADDRESS:
		return run_set_address(command[1]);
	case CMD_SET_BRIGHTNESS:
		run_set_leds(command + 1, MODULE_LENGTH);
		led_send_frame();
		return E_SUCCESS;
	case CMD_DUMP_CONFIG:
		return run_dump_config();
	case CMD_SAVE_CONFIG:
		save_config();
		return E_SUCCESS;
	default:
		return E_WRONGCOMMAND;
	}
}

/*
 * Runs the command pointed to by 'command'. How the command is interpreted
 * depends on which mode the LED board is in.
 *
 * Returns an error/success code.
 */
error_t run_command(uint8_t *command) {
#ifdef TRACE_COMMANDS
	debug_putchar(command[0]);
#endif

	if (my_mode == NORMAL_MODE) {
		return run_command_normal(command);
	} else {
		return run_command_config(command);
	}
}
